using System.Collections.Generic;
using System.Diagnostics;
using Antlr4.Runtime;
using Microsoft.Pc.Antlr;
using Microsoft.Pc.TypeChecker.AST.Declarations;

namespace Microsoft.Pc.TypeChecker.AST.States
{
    public class State : IPDecl
    {
        private readonly IDictionary<PEvent, IStateAction> actions = new Dictionary<PEvent, IStateAction>();

        public State(ParserRuleContext sourceNode, string name)
        {
            Debug.Assert(sourceNode is PParser.StateDeclContext);
            Name = name;
            SourceLocation = sourceNode;
        }

        public StateTemperature Temperature { get; set; }

        public bool IsStart { get; set; }
        public Function Entry { get; set; }
        public IEnumerable<KeyValuePair<PEvent, IStateAction>> AllEventHandlers => actions;

        public Function Exit { get; set; }
        public Machine OwningMachine { get; set; }
        public IStateContainer Container { get; set; }

        public IStateAction this[PEvent index] { get => actions[index]; set => actions[index] = value; }

        public string Name { get; }
        public ParserRuleContext SourceLocation { get; }

        public bool HasHandler(PEvent pEvent) { return actions.ContainsKey(pEvent); }
    }
}
